import numpy as np
import sys
import time
import datetime
import re
import cv2
from pyzm.helpers.Base import Base


from PIL import Image
import portalocker
import os

from pycoral.adapters import common
from pycoral.adapters import detect
from pycoral.utils.dataset import read_label_file
from pycoral.utils.edgetpu import make_interpreter
#from edgetpu.detection.engine import DetectionEngine

# Class to handle Yolo based detection


class Tpu(Base):

    def __init__(self, options={},logger=None ):
        Base.__init__(self,logger)
        self.classes = {}
        self.options = options
       #self.logger.Debug (1, 'UID:{} EUID:{}'.format( os.getuid(), os.geteuid()))

        #self.logger.Debug (4, 'TPU init params: {}'.format(options))
        
        self.processor='tpu'
        self.lock_maximum=int(options.get(self.processor+'_max_processes') or 1)
        self.lock_name='pyzm_uid{}_{}_lock'.format(os.getuid(),self.processor)
        self.lock_timeout = int(options.get(self.processor+'_max_lock_wait') or 120)
        self.disable_locks = options.get('disable_locks', 'no')
        if self.disable_locks == 'no':
            self.logger.Debug (2,f'portalock: max:{self.lock_maximum}, name:{self.lock_name}, timeout:{self.lock_timeout}')
            self.lock = portalocker.BoundedSemaphore(maximum=self.lock_maximum, name=self.lock_name,timeout=self.lock_timeout)
        self.is_locked = False
        self.model = None

        self.populate_class_labels()

    def acquire_lock(self):
        if self.disable_locks == 'yes':
            return
        if self.is_locked:
            self.logger.Debug (2, '{} portalock already acquired'.format(self.lock_name))
            return
        try:
            self.logger.Debug (2,'Waiting for {} portalock...'.format(self.lock_name))
            self.lock.acquire()
            self.logger.Debug (2,'Got {} portalock'.format(self.lock_name))
            self.is_locked = True

        except portalocker.AlreadyLocked:
            self.logger.Error ('Timeout waiting for {} portalock for {} seconds'.format(self.lock_name, self.lock_timeout))
            raise ValueError ('Timeout waiting for {} portalock for {} seconds'.format(self.lock_name, self.lock_timeout))

    def release_lock(self):
        if self.disable_locks == 'yes':
            return
        if not self.is_locked:
            self.logger.Debug (2, '{} portalock already released'.format(self.lock_name))
            return
        self.lock.release()
        self.is_locked = False
        self.logger.Debug (2,'Released portalock {}'.format(self.lock_name))


    def populate_class_labels(self):
        class_file_abs_path = self.options.get('object_labels')        
        for row in open(class_file_abs_path):
            # unpack the row and update the labels dictionary
            (classID, label) = row.strip().split(" ", maxsplit=1)
            self.classes[int(classID)] = label.strip()

    def get_classes(self):
        return self.classes

    def load_model(self):
        self.logger.Debug (1, '|--------- Loading TPU model from disk -------------|')
        start = datetime.datetime.now()

        #self.model = DetectionEngine(self.options.get('object_weights'))
        # Initialize the TF interpreter
        self.model = make_interpreter(self.options.get('object_weights'))

        self.model.allocate_tensors()

        diff_time = (datetime.datetime.now() - start)
        self.logger.Debug(
            1,'TPU initialization (loading model from disk) took: {}'
            .format(diff_time))
        
    def detect(self, image=None):
        Height, Width = image.shape[:2]
        img = image.copy()
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = Image.fromarray(img)

        if self.options.get('auto_lock',True):
            self.acquire_lock()

        try:
            if not self.model:
                self.load_model()

            self.logger.Debug(1,
                '|---------- TPU (input image: {}w*{}h) ----------|'
                .format(Width, Height))
            start = datetime.datetime.now()
            
            _, scale = common.set_resized_input(
                self.model, img.size, lambda size: img.resize(size, Image.ANTIALIAS))
            self.model.invoke()
            objs = detect.get_objects(self.model, float(self.options.get('object_min_confidence')), scale)

        
            #outs = self.model.detect_with_image(img, threshold=int(self.options.get('object_min_confidence')),
            #        keep_aspect_ratio=True, relative_coord=False)
            diff_time = (datetime.datetime.now() - start)

            if self.options.get('auto_lock',True):
                self.release_lock()
        except:
            if self.options.get('auto_lock',True):
                self.release_lock()
            raise

        self.logger.Debug(
            1,'Coral TPU detection took: {}'.format(diff_time))
      
        bbox = []
        labels = []
        conf = []

        for obj in objs:
           # box = obj.bbox.flatten().astype("int")
            bbox.append([
                    int(round(obj.bbox.xmin)),
                    int(round(obj.bbox.ymin)),
                    int(round(obj.bbox.xmax)),
                    int(round(obj.bbox.ymax))
                ])
            labels.append(self.classes[obj.id])
            conf.append(float(obj.score))

        self.logger.Debug(3,'Coral returning: {},{},{}'.format(bbox,labels,conf))
        return bbox, labels, conf



"""
Media
======
Generic media class that can read frames from either
a media file or via ZM APIs/index.php

"""

from pyzm.helpers.Base import Base
import requests
import cv2
import numpy as np
import imutils
from imutils.video import FileVideoStream
import time
import os
import random

class MediaStream(Base):
    def __init__(self, stream=None, type='video', options={}, logger=None):
        self.stream = stream
        self.type = type
        self.fvs = None
        self.next_frameid_to_read = 1
        self.last_frameid_read = 0
        self.options = options
        self.more_images_to_read = True
        self.frames_processed = 0
        self.frames_read = 0
        self.api = self.options.get('api')
        self.frame_set = []
        self.next_frame_set_index = 0
        self.orig_h_w = None
        self.resized_h_w = None        
        self.default_resize = 800
        self.is_deletable = False
        self.session = requests.Session()
        self.debug_filename = None 

              
        if logger:
            self.logger = logger
        elif self.options.get('api'):
            self.logger = options.get('api').get_logger() 

        _,ext= os.path.splitext(self.stream)
        if ext.lower() in ['.jpg', '.png', '.jpeg']:
            self.logger.Debug(1, '{} is a file'.format(self.stream))
            self.type = 'file'
            return
       
        self.start_frame = int(options.get('start_frame',1))
        self.frame_skip = int(options.get('frame_skip', 1))
        self.max_frames = int(options.get('max_frames', 0))
        self.contig_frames_before_error = int(options.get('contig_frames_before_error',5))
        self.frames_before_error = 0
        
        
        if self.stream.isnumeric(): 
        # assume it is an eid, in which case we 
        # need to access it via ZM API
            if not self.options.get('api'):
                self.logger.Error('API object not provided, cannot download event {}'.format(self.stream))
                raise ValueError('API object not provided, cannot download event {}'.format(self.stream))
            
            # we can either download the event, or access it via ZM
            # indirection 
            self.debug_filename = '{}-image'.format(self.stream)
            if self.options.get('download'):
                self.logger.Debug(1,'Downloading event:{}'.format(self.stream))
                es = self.api.events({'event_id': int(self.stream)})
                if not len(es.list()):
                    self.logger.Error('No such event {} found with API'.format(self.stream))
                    return
                e = es.list()[0]
                self.stream = e.download_video(dir=self.options.get('download_dir', '/tmp/'))
                self.is_deletable = True
                self.type = 'video'
            else:
                eid = self.stream
                self.next_frameid_to_read = int(self.start_frame)
                self.stream = '{}/index.php?view=image&eid={}'.format(self.api.get_portalbase(),  eid)
                #if self.options.get('resize') != 'no':
                #    self.stream += '&width={}'.format(self.options.get('resize', self.default_resize))

                #if self.api.get_auth() != '':
                #    self.stream += '&{}'.format(self.api.get_auth())

                self.logger.Debug(2,'Using URL {} for stream'.format(stream))
                self.type = 'image'
        
        
        if self.options.get('frame_set'):
            self.frame_set = self.options.get('frame_set').split(',')
            if 'alarm' in self.frame_set or 'snapshot' in self.frame_set:
                if not self.api or self.type == 'video':
                    # if we are not using ZM indirection, we cannot use 'alarm' 'snapshot' etc.
                
                    self.logger.Error ('You are using frame_types that require ZM indirection')
                    raise ValueError ('You are using frame_types that require ZM indirection')
            self.max_frames = len (self.frame_set)
            self.logger.Debug (2, 'We will only process frames: {}'.format(self.frame_set))
            self.next_frame_set_index = 0
            self.start_frame = self.frame_set[self.next_frame_set_index]
    
        if (self.type=='video'):
            self.logger.Debug (1, 'Starting video stream {}'.format(stream))
            f,_=os.path.splitext(os.path.basename(self.stream))
            self.debug_filename = '{}-image'.format(f)

            self.fvs = FileVideoStream(self.stream).start()
            time.sleep(1)
            self.logger.Debug (2, 'First load - skipping to frame {}'.format(self.start_frame))
            if self.frame_set:
                while self.fvs.more() and self.frames_read < int(self.frame_set[self.next_frame_set_index])-1:
                    self.fvs.read() 
                    self.frames_read +=1

            else:
                while self.fvs.more() and self.next_frameid_to_read < int(self.start_frame):
                    self.fvs.read()
                    self.next_frameid_to_read += 1

        else:
            self.logger.Debug (2,'No need to start streams, we are picking images from {}'.format(self.stream))

    def get_debug_filename(self):
        return self.debug_filename

    def image_dimensions(self):
        return {
            'original': self.orig_h_w,
            'resized': self.resized_h_w
        }

    def get_last_read_frame(self):
        return self.last_frameid_read

    def more(self):

        if self.type == 'file':
            return self.more_images_to_read

        if (self.frame_set):
            return self.next_frame_set_index < len(self.frame_set)

        if (self.frames_processed >= self.max_frames):
            self.logger.Debug(1, 'Bailing as we have read {} frames'.format(self.max_frames))
            return False
        else:
            if (self.type == 'video'):
                return self.fvs.more()
            else:
                return self.more_images_to_read

    def stop(self):
        if (self.type == 'video'):
            self.fvs.stop()
        if self.is_deletable and self.options.get('delete_after_analyze') == 'yes':
            try:
                os.remove(self.stream)
                self.logger.Debug(2,'Deleted {}'.format(self.stream))
            except Exception as e:
                self.logger.Error (f'Could not delete file(s):{e}')

    def read(self):
        if (self.type == 'file'):
            frame = cv2.imread(self.stream)
            self.last_frame_id_read = 1
            self.orig_h_w = frame.shape[:2]
            self.frames_processed += 1
            self.more_images_to_read = False
            if self.options.get('resize') != 'no':
                frame = imutils.resize(frame,width=self.options.get('resize', self.default_resize))
            self.resized_h_w = frame.shape[:2]  
            return frame


        if (self.type == 'video'):
            while True:
                frame = self.fvs.read()
                self.frames_read +=1

                if frame is None:
                    self.frames_before_error +=1
                    if (self.frames_before_error >= self.contig_frames_before_error):
                        self.logger.Error ('Error reading frames')
                        return
                    else:
                        self.logger.Debug (1, 'Error reading frame: {} of max {} contiguous errors'.format(self.frames_before_error, self.contig_frames_before_error))
                        continue

                self.frames_before_error = 0
                self.orig_h_w = frame.shape[:2]
                if self.frame_set and (self.frames_read != int(self.frame_set[self.next_frame_set_index])):
                    continue
                    # At this stage we are at the frame to read
                if self.frame_set:
                    self.last_frameid_read = self.frame_set[self.next_frame_set_index]
                    self.next_frame_set_index += 1
                    if self.next_frame_set_index < len (self.frame_set):
                        self.logger.Debug (4,"Now moving to frame: {}".format(self.frame_set[self.next_frame_set_index])) 

                else:
                    self.last_frameid_read = self.next_frameid_to_read
                    self.next_frameid_to_read +=1
                    if (self.last_frameid_read - 1 ) % self.frame_skip:
                        self.logger.Debug(5,'Skipping frame {}'.format(self.last_frameid_read))
                        continue    
                    
                self.logger.Debug(2, 'Processing frame:{}'.format(self.last_frameid_read))
                self.frames_processed += 1
                break
            #print ('******************************* RESIZE:{}'.format(self.options.get('resize')))
            if self.options.get('resize') != 'no':
                frame = imutils.resize(frame,width=self.options.get('resize', self.default_resize))
            self.resized_h_w = frame.shape[:2]

            if self.options.get('save_frames') and self.debug_filename:
                d = self.options.get('save_frames_dir','/tmp')
                fname = '{}/{}-{}.jpg'.format(d,self.debug_filename,self.last_frameid_read) 
                self.logger.Debug (4, 'Saving image to {}'.format(fname))
                cv2.imwrite(fname,frame)        

            return frame
        
        else: # image
            if self.frame_set:
                if self.next_frame_set_index >= len (self.frame_set):
                    self.logger.Debug (1,'Reached end of frame_set')
                    self.more_images_to_read = False
                    self.next_frameid_to_read = 0
                    return None 
                url = '{}&fid={}'.format(self.stream,self.frame_set[self.next_frame_set_index])
            else:
                url = '{}&fid={}'.format(self.stream,self.next_frameid_to_read)
            
            self.logger.Debug (3, 'Reading {}'.format(url))
            response = None
            try:
                if self.api:
                    attempts = 1
                    max_attempts = self.options.get('max_attempts',1)
                    sleep_time = self.options.get('sleep_between_attempts',3)
                    while attempts <= max_attempts:
                        try:
                            response = self.api._make_request(url)
                        except Exception as e:
                            err_msg = '{}'.format(e)
                            if err_msg != 'BAD_IMAGE':
                                break
                            self.logger.Debug (2,'Failed attempt:{} of {}'.format(attempts, max_attempts))
                            attempts += 1
                            if sleep_time and attempts <=max_attempts:
                                self.logger.Debug (2,'Sleeping for {} seconds before retry'.format(sleep_time))
                                time.sleep(sleep_time)
                        else: # request worked, no need to retry
                            break
                else:
                    response = self.session.get(url)
            except Exception as e:
                if self.frame_set:
                    self.next_frame_set_index += 1
                    if self.next_frame_set_index < len(self.frame_set):
                        self.logger.Error('Error reading frame: {}, but moving to next frame_set'.format(url))
                        self.last_frameid_read = self.frame_set[self.next_frame_set_index]
                        return self.read()
                    else:
                        self.logger.Error('Error reading frame: {} and no more frames to read'.format(url))
                        self.more_images_to_read = False
                        self.next_frameid_to_read = 0
                        return None 
                        
            if self.frame_set:
                self.last_frameid_read = self.frame_set[self.next_frame_set_index]
                self.next_frame_set_index += 1
            else:
                self.last_frameid_read = self.next_frameid_to_read
                self.next_frameid_to_read +=self.frame_skip

            self.frames_processed +=1 
            if response and response.status_code == 200: # and random.randint(0,5)==1:
                self.frames_before_error = 0
                try:
                    img = np.asarray(bytearray(response.content), dtype='uint8')
                    img = cv2.imdecode (img, cv2.IMREAD_COLOR)
                    self.orig_h_w =  img.shape[:2]
                    #print ('******************************* RESIZE:{}'.format(self.options.get('resize')))
                    if self.options.get('resize') != 'no':
                        img = imutils.resize(img,width=self.options.get('resize', self.default_resize))
                    self.resized_h_w = img.shape[:2]

                    if self.options.get('save_frames') and self.debug_filename:
                        d = self.options.get('save_frames_dir','/tmp')
                        fname = '{}/{}-{}.jpg'.format(d,self.debug_filename,self.last_frameid_read) 
                        self.logger.Debug (4, 'Saving image to {}'.format(fname))
                        cv2.imwrite(fname,img)        

                    return img
                except Exception as e:
                    self.logger.Error ('Could not retrieve url {}: {}'.format(url,e))
                    return None
          
            else: # response code not 200
                self.frames_before_error +=1
                if (self.frames_before_error >= self.contig_frames_before_error):
                    self.logger.Error ('Error reading frames')
                    return None
                else:
                    self.logger.Debug (1, 'Error reading frame: {} of max {} contiguous errors'.format(self.frames_before_error, self.contig_frames_before_error))
                    return self.read()
                self.more_images_to_read = False
                self.next_frameid_to_read = 0
                return None 



import numpy as np

import pyzm.ml.face_train as train
import face_recognition
import dlib

import sys
import os
import cv2
import pickle
from sklearn import neighbors
import imutils
import math
import uuid
import time
import datetime
from pyzm.helpers.Base import Base
# Class to handle face recognition
import portalocker
import re
from pyzm.helpers.Media import MediaStream
import imutils

class Face(Base):
    def __init__(self, logger=None, options={},upsample_times=1, num_jitters=0, model='hog'):
        super().__init__(logger)
        #self.logger.Debug (4, 'Face init params: {}'.format(options))

        if dlib.DLIB_USE_CUDA and dlib.cuda.get_num_devices() >=1 :
            self.processor = 'gpu'
        else:
            self.processor = 'cpu'

        self.logger.Debug(
            1,'Initializing face recognition with model:{} upsample:{}, jitters:{}'
            .format(model, upsample_times, num_jitters))

        self.disable_locks = options.get('disable_locks', 'no')

        self.upsample_times = upsample_times
        self.num_jitters = num_jitters
        if options.get('face_model'):
            self.face_model = options.get('face_model')
        else:
            self.face_model = model
       
        self.knn = None
        self.options = options
        self.is_locked = False

        self.lock_maximum=int(options.get(self.processor+'_max_processes') or 1)
        self.lock_timeout = int(options.get(self.processor+'_max_lock_wait') or 120)
        
        #self.lock_name='pyzm_'+self.processor+'_lock'
        self.lock_name='pyzm_uid{}_{}_lock'.format(os.getuid(),self.processor)
        if self.disable_locks == 'no':
            self.logger.Debug (2,f'portalock: max:{self.lock_maximum}, name:{self.lock_name}, timeout:{self.lock_timeout}')
            self.lock = portalocker.BoundedSemaphore(maximum=self.lock_maximum, name=self.lock_name,timeout=self.lock_timeout)
            

        if self.options.get('face_detection_framework') != 'dlib' and self.options.get('face_recognition_framework') != 'dlib':
            raise ValueError ('Error: As of now,only dlib is supported for face detection and recognition. Unkown {}/{}'.format(self.options.get('face_detection_framework'),self.options.get('face_recognition_framework')))

        encoding_file_name = self.options.get('known_images_path') + '/faces.dat'
        try:
            if (os.path.isfile(self.options.get('known_images_path') +
                               '/faces.pickle')):
                # old version, we no longer want it. begone
                self.logger.Debug(
                    1,'removing old faces.pickle, we have moved to clustering')
                os.remove(self.options.get('known_images_path') + '/faces.pickle')
        except Exception as e:
            self.logger.Error('Error deleting old pickle file: {}'.format(e))

        # to increase performance, read encodings from  file
        if (os.path.isfile(encoding_file_name)):
            self.logger.Debug(
                1,'pre-trained faces found, using that. If you want to add new images, remove: {}'
                .format(encoding_file_name))

            #self.known_face_encodings = data["encodings"]
            #self.known_face_names = data["names"]
        else:
            # no encodings, we have to read and train
            self.logger.Debug(
                1,'trained file not found, reading from images and doing training...'
            )
            self.logger.Debug(
                1,'If you are using a GPU and run out of memory, do the training using zm_train_faces.py. In this case, other models like yolo may already take up a lot of GPU memory'
            )

            train.FaceTrain(options=self.options).train()
        try:
            with open(encoding_file_name, 'rb') as f:
                self.knn = pickle.load(f)
        except Exception as e:
            self.logger.Error ('Error loading KNN model: {}'.format(e))


    def acquire_lock(self):
        if self.disable_locks=='yes':
            return
        if self.is_locked:
            self.logger.Debug (2, '{} portalock already acquired'.format(self.lock_name))
            return
        try:
            self.logger.Debug (2,f'Waiting for {self.lock_name} portalock...')
            self.lock.acquire()
            self.logger.Debug (2,f'Got {self.lock_name} lock...')
            self.is_locked = True

        except portalocker.AlreadyLocked:
            self.logger.Error ('Timeout waiting for {} portalock for {} seconds'.format(self.lock_name, self.lock_timeout))
            raise ValueError ('Timeout waiting for {} portalock for {} seconds'.format(self.lock_name, self.lock_timeout))


    def release_lock(self):
        if self.disable_locks=='yes':
            return
        if not self.is_locked:
            self.logger.Debug (1, '{} portalock already released'.format(self.lock_name))
            return
        self.lock.release()
        self.is_locked = False
        self.logger.Debug (1,'Released {} portalock'.format(self.lock_name))

    def get_classes(self):
        if self.knn:
            return self.knn.classes_
        else:
            return []

    def _rescale_rects(self, a):
        rects = []
        for (left, top, right, bottom) in a:
            top *= 4
            right *= 4
            bottom *= 4
            left *= 4
            rects.append([left, top, right, bottom])
        return rects

    

    def detect(self, image):

        Height, Width = image.shape[:2]
        self.logger.Debug(
            1,'|---------- Face recognition (input image: {}w*{}h) ----------|'.
            format(Width, Height))

        downscaled =  False
        upsize_xfactor = None
        upsize_yfactor = None
        max_size = self.options.get('max_size', Width)
        old_image = None

        if Width > max_size:
            downscaled = True
            self.logger.Debug (2, 'Scaling image down to max size: {}'.format(max_size))
            old_image = image.copy()
            image = imutils.resize(image,width=max_size)
            newHeight, newWidth = image.shape[:2]
            upsize_xfactor = Width/newWidth
            upsize_yfactor = Height/newHeight

        labels = []
        classes = []
        conf = []

        # Convert the image from BGR color (which OpenCV uses) to RGB color (which face_recognition uses)
        rgb_image = image[:, :, ::-1]
        #rgb_image = image

        # Find all the faces and face encodings in the target image
        #prin (self.options)
        if self.options.get('auto_lock',True):
            self.acquire_lock()

        start = datetime.datetime.now()
        face_locations = face_recognition.face_locations(
            rgb_image,
            model=self.face_model,
            number_of_times_to_upsample=self.upsample_times)

        diff_time = (datetime.datetime.now() - start)
        self.logger.Debug(1,'Finding faces took {}'.format(diff_time))

        start = datetime.datetime.now()
        face_encodings = face_recognition.face_encodings(
            rgb_image,
            known_face_locations=face_locations,
            num_jitters=self.num_jitters)
        
        if self.options.get('auto_lock',True):
            self.release_lock()

        diff_time = (datetime.datetime.now() - start)
        self.logger.Debug(
            1,'Computing face recognition distances took {}'.format(
                diff_time))

        if not len(face_encodings):
            return [], [], []

        # Use the KNN model to find the best matches for the test face
      
        start = datetime.datetime.now()

        if self.knn:
            closest_distances = self.knn.kneighbors(face_encodings, n_neighbors=1)
            are_matches = [
                closest_distances[0][i][0] <= int(self.options.get('face_recog_dist_threshold',0.6))
                for i in range(len(face_locations))
                
            ]
            prediction_labels = self.knn.predict(face_encodings)

        else:
            # There were no faces to compare
            # create a set of non matches for each face found
            are_matches = [False] * len(face_locations)
            prediction_labels = [''] * len(face_locations)
            self.logger.Debug (1,'No faces to match, so creating empty set')

        diff_time = (datetime.datetime.now() - start)
        self.logger.Debug(
            1,'Matching recognized faces to known faces took {}'.
            format(diff_time))

        matched_face_names = []
        matched_face_rects = []

        if downscaled:
            self.logger.Debug (2,'Scaling image back up to {}'.format(Width))
            image = old_image
            new_face_locations = []
            for loc in face_locations:
                a,b,c,d=loc
                a = round(a * upsize_yfactor)
                b = round(b * upsize_xfactor)
                c = round(c * upsize_yfactor)
                d = round(d * upsize_xfactor)
                new_face_locations.append((a,b,c,d))
            face_locations = new_face_locations


        for pred, loc, rec in zip(prediction_labels,
                                  face_locations, are_matches):
            label = pred if rec else self.options.get('unknown_face_name', 'unknown')
            if not rec and self.options.get('save_unknown_faces') == 'yes':
                h, w, c = image.shape
                x1 = max(loc[3] - int(self.options.get('save_unknown_faces_leeway_pixels',0)),0)
                y1 = max(loc[0] - int(self.options.get('save_unknown_faces_leeway_pixels',0)),0)
                x2 = min(loc[1] + int(self.options.get('save_unknown_faces_leeway_pixels',0)), w)
                y2 = min(loc[2] + int(self.options.get('save_unknown_faces_leeway_pixels',0)),h)
                #print (image)
                crop_img = image[y1:y2, x1:x2]
                # crop_img = image
                timestr = time.strftime("%b%d-%Hh%Mm%Ss-")
                unf = self.options.get('unknown_images_path') + '/' + timestr + str(
                    uuid.uuid4()) + '.jpg'
                self.logger.Info(
                    'Saving cropped unknown face at [{},{},{},{} - includes leeway of {}px] to {}'
                    .format(x1, y1, x2, y2,
                            self.options.get('save_unknown_faces_leeway_pixels'), unf))
                cv2.imwrite(unf, crop_img)

            
            matched_face_rects.append([loc[3], loc[0], loc[1], loc[2]])
            matched_face_names.append(label)
            #matched_face_names.append('face:{}'.format(label))
            conf.append(1)

        self.logger.Debug(3,f'FACE:Returning: {matched_face_rects}, {matched_face_names}, {conf}')
        return matched_face_rects, matched_face_names, conf
